!---------------------------------------------------------------------------------------
! PREP_SYNDATA: mod_perturb
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_perturb
!!
!! Purpose:    Module for randomly perturbing the prior and observation vectors.
!!
!---------------------------------------------------------------------------------------

module mod_perturb

  use mod_var
  use mod_settings

  implicit none
  private

  public :: perturb_flux, perturb_obs, randnum, random_normal

  contains

  ! --------------------------------------------------
  ! perturb_flux
  ! --------------------------------------------------
  !> perturb_flux
  !!
  !! Purpose:  Adds random perturbation to the prior
  !!           state vector according to error 
  !!           characteristics of the error covariance
  !!           matrix.
  !!
  ! --------------------------------------------------

  subroutine perturb_flux(config, files, state)

    use mod_var
    use mod_settings

    type (config_t),                  intent (in)     :: config
    type (files_t),                   intent (in)     :: files
    real, dimension(nvar),            intent (in out) :: state

    real(kind=8), dimension(neig)                     :: tmp
    real(kind=8), dimension(ndvar)                    :: tmp1
    real(kind=8), dimension(nvar)                     :: perturb
    real(kind=8), dimension(ntstate,ntstate)          :: sqcort, work
    real, dimension(nvar)                             :: rvec
    integer                                           :: seed
    integer                                           :: info
    integer                                           :: i, j, k, it, jt

    ! square root of cort
    ! use Cholesky factorization
    work = cort
    call dpotrf('L',ntstate,work,ntstate,info)
    if ( info.gt.0 ) then
      write(logid,*) 'ERROR perturb_flux: lower Cholesky triangle'
      stop
    endif
    call dpotri('L',ntstate,work,ntstate,info)
    if ( info.gt.0 ) then
      write(logid,*) 'ERROR perturb_flux: Cholesky inverse'
      stop
    endif
    ! returns lower ('L') triangle of inverse of work
    sqcort = work
    do i = 1, ntstate
      do j = 1, i
        sqcort(j,i) = work(i,j)
      end do
    end do 

    ! random number vector
    seed = config%seed
    call randnum(nvar, seed, rvec)

    ! perturb = B^(1/2)r 
    perturb(:) = 0d0
    do jt = 1, ntstate
      do k = 1, ndvar
        do j = 1, neig
          tmp(j) = evecs(k,j)*sqrt(evals(j))
        end do
        do i = 1, ndvar
          tmp1(i) = dot_product(tmp,evecs(i,:))
        end do
        do it = 1, ntstate
          perturb((jt-1)*ndvar+k) = perturb((jt-1)*ndvar+k) + sqcort(jt,it) * &
                                     dot_product(tmp1,dble(rvec((it-1)*ndvar+1:it*ndvar)))
        end do
      end do
    end do
    write(logid,*) 'perturb_flux: min/max perturbation = ',minval(perturb), maxval(perturb)

    ! add perturbation to state vector 
    state(:) = state(:) + real(perturb(:),kind=4)

    ! testing
    open(100,file=trim(files%path_output)//'pert_state.txt',status='replace',action='write')
    do i = 1, nvar
      write(100,fmt='(F14.8)') perturb(i)
    end do
    close(100)
    open(100,file=trim(files%path_output)//'state_pert.txt',status='replace',action='write')
    do i = 1, nvar
      write(100,fmt='(F14.8)') state(i)
    end do
    open(100,file=trim(files%path_output)//'rvec.txt',status='replace',action='write')
    do i = 1, nvar
      write(100,fmt='(F10.8)') rvec(i)
    end do
    close(100)

    if ( config%lognormal ) then
      ! lognormal
      state(:) = exp(state)
    endif

  end subroutine perturb_flux

  ! --------------------------------------------------
  ! perturb_obs
  ! --------------------------------------------------
  !> perturb_obs
  !!
  !! Purpose:  Adds random perturbation to the 
  !!           observation vector using the error
  !!           characteristics of the observation 
  !!           space errors.
  !!
  ! --------------------------------------------------

  subroutine perturb_obs(config, files, obs, obserr)

    use mod_var
    use mod_settings

    type (config_t),         intent (in)     :: config
    type (files_t),          intent (in)     :: files
    real, dimension(maxobs), intent (in out) :: obs 
    real, dimension(maxobs), intent (in)     :: obserr 

    real, dimension(nobs)                    :: perturb
    real, dimension(nobs)                    :: rvec
    integer                                  :: i, seed

    ! random number vector
    seed = config%seed
    call randnum(nobs, seed, rvec)

    ! perturb = R^(1/2)r
    perturb(:) = 0.
    do i = 1, nobs
      perturb(i) = obserr(i)*rvec(i)
    end do
 
    ! obs = obs + perturb
    obs(1:nobs) = obs(1:nobs) + perturb(:)

    ! testing
    open(100,file=trim(files%path_output)//'pert_obs.txt',status='replace',action='write')
    do i = 1, nobs
      write(100,fmt='(F10.4)') perturb(i)
    end do
    close(100)
    open(100,file=trim(files%path_output)//'obs_pert.txt',status='replace',action='write')
    do i = 1, nobs
      write(100,fmt='(F10.4)') obs(i)
    end do
    close(100)

  end subroutine perturb_obs

  ! --------------------------------------------------
  ! randnum
  ! --------------------------------------------------
  !> randnum
  !!
  !! Purpose:  Generates a vector of random numbers
  !!           with a mean of zero and SD of one.
  !!
  ! --------------------------------------------------

  subroutine randnum(num, seed, rvec)

    integer,               intent (in)     :: num
    integer,               intent (in)     :: seed
    real, dimension(num),  intent (in out) :: rvec

    integer                                :: ssize, i
    real                                   :: ri
    integer, dimension(:), allocatable     :: seedvec

    ! generate vector of random numbers
    call random_seed
    call random_seed(size=ssize)
    allocate(seedvec(ssize))
    do i = 1, ssize
      seedvec(i) = seed
    end do
    call random_seed(put=seedvec(1:ssize))
    do i = 1, num
!      call random_number(ri)
      ri = random_normal()
      rvec(i) = ri
    end do
    deallocate(seedvec)

    ! note: normalization not needed with random_normal function!
    ! normalize random numbers
!    rvec = rvec*num/sum(rvec)
 
    ! mean of zero
!    rvec = rvec - 1.

  end subroutine randnum

  ! --------------------------------------------------
  ! random_normal
  ! --------------------------------------------------
  !> random_normal
  !!
  !! Adapted from the following Fortran 77 code
  !!      ALGORITHM 712, COLLECTED ALGORITHMS FROM ACM.
  !!      THIS WORK PUBLISHED IN TRANSACTIONS ON 
  !!      MATHEMATICAL SOFTWARE, VOL. 18, NO. 4, 
  !!      DECEMBER, 1992, PP. 434-435.
  !!
  !! Purpose:  
  !!  The function random_normal() returns a normally 
  !!  distributed pseudo-random number with zero mean 
  !!  and unit variance.
  !!
  !!  The algorithm uses the ratio of uniforms method 
  !!  of A.J. Kinderman and J.F. Monahan augmented 
  !!  with quadratic bounding curves.
  !!
  ! --------------------------------------------------

  real function random_normal()
 
    implicit none

    real :: s = 0.449871, t = -0.386595, a = 0.19600, b = 0.25472,           &
            r1 = 0.27597, r2 = 0.27846, u, v, x, y, q, half = 0.5

    ! Generate P = (u,v) uniform in rectangle enclosing acceptance region

    do
      call random_number(u)
      call random_number(v)
      v = 1.7156 * (v - half)

      ! Evaluate the quadratic form
      x = u - s
      y = abs(v) - t
      q = x**2 + y*(a*y - b*x)

      ! Accept P if inside inner ellipse
      if (q < r1) exit
      ! Reject P if outside outer ellipse
      if (q > r2) cycle
      ! Reject P if outside acceptance region
      if (v**2 < -4.0*log(u)*u**2) exit
    end do

    ! Return ratio of P's coordinates as the normal deviate
    random_normal = v/u
    return

  end function random_normal

  ! --------------------------------------------------


end module mod_perturb



