!---------------------------------------------------------------------------------------
! FLEXINVERT: error_cov
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> error_cov
!!
!! Purpose:    Calculates the prior error covariance matrix for the state variables.
!!
!! Interface:  
!!
!!    Inputs
!!             config  -  configuration settings data structure
!!             files   -  file data structure
!!             states  -  state vector data structure
!!             covar   -  error covariance data structure
!!             corr    -  spatial correlation matrix
!!             xerr    -  uncertainties for state variables
!!
!!    Externals
!!             dsyevr (LAPACK routine)
!!             alloc_covar
!!
!---------------------------------------------------------------------------------------

subroutine error_cov(config, files, states, covar, corr, xerr)

  use mod_var
  use mod_settings
  use mod_states
  use mod_covar

  implicit none

  type (config_t),            intent (in)     :: config
  type (files_t),             intent (in)     :: files
  type (states_t),            intent (in out) :: states
  type (covar_t),             intent (in out) :: covar
  real, dimension(nbox,nbox), intent (in)     :: corr
  real, dimension(ndvar),     intent (in)     :: xerr

  character(len=max_path_len)                 :: filename
  character(len=max_path_len)                 :: rowfmt
  integer                                     :: i, j, k, it, jt, nt
  real                                        :: toterr
  real                                        :: erscalar
  real                                        :: covsum
  real(kind=8), dimension(ndvar,ndvar)        :: cov
  real(kind=8), dimension(ntstate)            :: tmp
  real(kind=8), dimension(nvar)               :: var
  real, dimension(ntstate)                    :: tsteps
  real, dimension(ntstate,ntstate)            :: tmat, dtmat
  real(kind=8), dimension(ntstate,ntstate)    :: cort, icort, sqcort, isqcort
  integer                                     :: lwork, liwork
  integer, dimension(:), allocatable          :: isuppz
  integer, dimension(:), allocatable          :: iwork
  real(kind=8), dimension(:), allocatable     :: work
  real(kind=8), dimension(:,:), allocatable   :: zwork
  real(kind=8), dimension(:), allocatable     :: evals
  real(kind=8), dimension(:,:), allocatable   :: evecs
  integer                                     :: ierr, info, neval, ind
  real, parameter                             :: smallnum=1.e-15
  integer                                     :: timei, timef, countrate

  ! calculate error covariance 
  ! --------------------------

  ! spatial error covariance one time step
  cov(:,:) = 0d0
  covsum = 0.
  ! sub-daily resolution (no correlation between sub-daily time steps)
  do nt = 1, ndt
    do j = 1, nbox
      do i = 1, nbox
        cov((nt-1)*nbox+i,(nt-1)*nbox+j) = dble(corr(i,j))*dble(xerr((nt-1)*nbox+i))*dble(xerr((nt-1)*nbox+j))
        covsum = covsum + real(cov((nt-1)*nbox+i,(nt-1)*nbox+j),kind=4)*area_box(j)*area_box(i)
      end do
    end do
  end do
  covsum = covsum/real(ndt)**2

  if ( .not.config%lognormal ) then
    ! scale to total error of nested domain (unit Tg/y)
    toterr = sqrt(covsum)*3600.*24.*365./1.e9/numscale
    write(logid,*) 'Total error: ',toterr
    erscalar = config%globerr/toterr
    if ( config%globerr > 0. ) then
      cov = cov * dble(erscalar)**2
      write(logid,*) 'Total error scaled by: ',erscalar
    end if
  endif

  ! calculate eigendecomposition of cov
  ! -----------------------------------

  ! call to dsyevr to find only ILth to IUth eigenvalues and vectors 
  ! thus reducing time needed for eigen decomposition for large B
  ! IL = index of smallest eigenvalue to be found
  ! IU = index of largest eigenvalue to be found
  ! where 1 <= IL <= IU <=N
  ! IU = N and IL = max(1, N - f*N) where f is fraction of eigenvalues to keep

  ! B = U*D*U^(T)
  lwork = 26*ndvar
  liwork = 10*ndvar
  allocate( work(lwork), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( iwork(liwork), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( isuppz(2*ndvar), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( evecs(ndvar,ndvar), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( evals(ndvar), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( zwork(ndvar,ndvar), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  zwork = cov

  ! arguments to dsyevr
  ! jobsz = N : N only eigenvalues, V both
  ! range = A : A all eigenvalues, V values in interval (VL,VU], I from ILth to IUth values
  ! uplo = L  : lower triangle of matrix A stored
  ! n = ndvar : order of matrix A
  ! matrix = zwork : matrix A 
  ! lda = ndvar : leading dimension of A
  ! vl = used if range is V : lower bound of interval to search for eigenvalues
  ! vu = used if range is V : upper bound of interval to search for eigenvalues
  ! il = used if range is I : index of smallest eigenvalue to be found
  ! iu = used if range is I : index of largest eigenvalue to be found
  ! absol = 0     : absolute tolerance for eigenvalues
  ! m = neval     : number of eigenvalues found
  ! w = evals     : eigenvalues in ascending order
  ! z = evecs     : found eigenvectors as columns
  ! ldz = ndvar   : leading dimension of z
  ! isuppz = isuppz : matrix of indices to non-zero values in z
  ! work = temporary working array
  ! lwork = dimension of work
  ! iwork = temporary working array
  ! liwork = dimension of iwork
  ! info : 0 success, < 0 if info = -i ith argument had an illegal value, > 0 internal error
 
  ! first calculate only eigenvalues
  call system_clock(timei, countrate)
  call dsyevr('N','A','L',ndvar,zwork,ndvar,1.,1.,1,1,0.,neval,evals,evecs,ndvar,isuppz,work,lwork,iwork,liwork,info)
  if ( info.ne.0 ) then
    write(logid,*) 'ERROR error_cov: eigenvalue decomposition'
    stop
  endif
  call system_clock(timef, countrate)
  print*, 'error_cov: time to calculate eigenvalues (s) = ',(timef-timei)/countrate

  ! now calculate eigenvalues in range IU = N and IL = max(1, N - f*N) 
  ! where f is fraction of eigenvalues to keep
  ! and their corresponding eigenvectors
  ind = minloc(evals, dim=1, mask=evals.ge.(config%trunc*maxval(evals)))
  print*, 'error_cov: ind, evals(ind) = ',ind, evals(ind)
  call system_clock(timei, countrate)
  ! because on exit the lower triangle (if UPLO='L') or the upper triangle 
  ! (if UPLO='U') of A, including the diagonal, is destroyed
  ! need to reassign matrix A
  zwork = cov
  evals(:) = 0.
  evecs(:,:) = 0.
  call dsyevr('V','I','L',ndvar,zwork,ndvar,1.,1.,ind,ndvar,0.,neval,evals,evecs,ndvar,isuppz,work,lwork,iwork,liwork,info)
  if ( info.ne.0 ) then
    write(logid,*) 'ERROR error_cov: eigenvalue decomposition'
    stop
  endif
  call system_clock(timef, countrate)
  print*, 'error_cov: time to calculate eigenvalues and eigenvectors (s) = ',(timef-timei)/countrate
  print*, 'error_cov: neval = ',neval

  ! truncate eigenvalues
  ! eigenvalues are sorted in ascending order
  neig = neval
  write(logid,*) 'Number of retained eigenvalues: ',neig

  call alloc_covar(covar)
  ! eigenvalues in ascending order
  ! covar%evals = evals(ind:ndvar)
  covar%evals = evals(1:neig)
  ! calculated eigenvectors in columns 1:neig
  covar%evecs = evecs(:,1:neig)

  deallocate(work)
  deallocate(zwork)
  deallocate(iwork)
  deallocate(isuppz)
  deallocate(evals)
  deallocate(evecs)
 
  ! calculate temporal correlation
  ! ------------------------------

  do i = 1, ntstate
    tsteps(i) = i * statres
  end do

  tmat = matmul(reshape(tsteps,(/ntstate,1/)),reshape((tsteps/tsteps),(/1,ntstate/)))
  dtmat = abs(tmat - transpose(tmat))

  do i = 1, ntstate
    do j = i, ntstate
      cort(i,j) = exp(-1d0*dble(dtmat(i,j)/config%sigmatime))
      cort(j,i) = exp(-1d0*dble(dtmat(j,i)/config%sigmatime))
    end do
  end do

  ! calculate eigendecomposition of cort
  ! ------------------------------------

  ! cort = U*D*U^(T)
  lwork = 26*ntstate
  liwork = 10*ntstate
  allocate( work(lwork), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( iwork(liwork), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( isuppz(2*ntstate), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( evecs(ntstate,ntstate), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( evals(ntstate), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'
  allocate( zwork(ntstate,ntstate), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR error_cov: not enough memory'

  zwork = cort
  call dsyevr('V','A','L',ntstate,zwork,ntstate,1.,1.,1,1,0.,neval,evals,evecs,ntstate,isuppz,work,lwork,iwork,liwork,info)
  if ( info.ne.0 ) then
    write(logid,*) 'ERROR error_cov: eigenvalue decomposition'
    stop
  endif

  ! inverse of cort
  icort(:,:) = 0d0
  do i = 1, ntstate
    do j = 1, ntstate
      tmp(j) = evecs(i,j)*1d0/evals(j)
    end do
    icort(i,:) = matmul(tmp,transpose(evecs))
  end do

  ! square root of cort
  sqcort(:,:) = 0d0
  do i = 1, ntstate
    do j = 1, ntstate
      tmp(j) = evecs(i,j)*sqrt(evals(j))
    end do
    sqcort(i,:) = matmul(tmp,transpose(evecs))
  end do

  ! inverse of square root of cort
  isqcort(:,:) = 0d0
  do i = 1, ntstate
    do j = 1, ntstate
      tmp(j) = evecs(i,j)*1d0/sqrt(evals(j))
    end do
    isqcort(i,:) = matmul(tmp,transpose(evecs))
  end do

  covar%cort = cort
  covar%icort = icort
  covar%sqcort = sqcort
  covar%isqcort = isqcort

  deallocate(work)
  deallocate(zwork)
  deallocate(iwork)
  deallocate(isuppz)
  deallocate(evals)
  deallocate(evecs)

  ! prior error
  ! -----------

  ! include cross-correlations
  var(:) = 0d0
  do jt = 1, ntstate
    do k = 1, ndvar
!      do it = 1, ntstate
!       var((jt-1)*ndvar+k) = var((jt-1)*ndvar+k) + covar%cort(jt,it)*sum(cov(k,:))
       var((jt-1)*ndvar+k) = covar%cort(jt,jt)*cov(k,k)
!      end do
    end do
  end do

  ! uncertainty for flux variables
  do i = 1, npvar
    states%pxerr0(i) = real(sqrt(var(i)),kind=4)
  end do

  ! uncertainty for scalars of initial mixing ratio
  if ( config%opt_cini ) then
    do i = 1, ntcini*ncini
      if ( config%lognormal ) then
        ! convert to error of z
        states%pxerr0(npvar+i) = log(1.+config%cinierr)
      else
        states%pxerr0(npvar+i) = config%cinierr
      endif
    end do
  endif

  ! save output
  ! -----------

  filename = trim(files%path_output)//'evals.txt'
  open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
  do i = 1, neig
    write(100,fmt='(E11.4)') covar%evals(i)
  end do
  close(100)
  filename = trim(files%path_output)//'evecs.txt'
  write(rowfmt,'(A,I6,A)') '(',neig,'(E11.4,1X))'
  open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
  do i = 1, ndvar
    write(100,fmt=rowfmt) covar%evecs(i,:)
  end do
  close(100)
  filename = trim(files%path_output)//'cort.txt'
  write(rowfmt,'(A,I6,A)') '(',ntstate,'(E11.4,1X))'
  open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
  do i = 1, ntstate
    write(100,fmt=rowfmt) covar%cort(i,:)
  end do
  close(100)
  filename = trim(files%path_output)//'sqcort.txt'
  write(rowfmt,'(A,I6,A)') '(',ntstate,'(E11.4,1X))'
  open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
  do i = 1, ntstate
    write(100,fmt=rowfmt) covar%sqcort(i,:)
  end do
  close(100)

  ! verbose output
  ! --------------

  if ( config%verbose ) then
    filename = trim(files%path_output)//'icort.txt'
    write(rowfmt,'(A,I6,A)') '(',ntstate,'(E11.4,1X))'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do i = 1, ntstate
      write(100,fmt=rowfmt) covar%icort(i,:)
    end do
    close(100)
    filename = trim(files%path_output)//'pxerr0.txt'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do i = 1, nvar
      write(100,fmt='(E11.4)') states%pxerr0(i)
    end do
    close(100)
    filename = trim(files%path_output)//'correl.txt'
    write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do i = 1, nbox
      write(100,fmt=rowfmt) corr(i,:)
    end do
    close(100)
  endif

end subroutine error_cov

